/*
 * Copyright (C) 2009-2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster;

import org.eiichiro.jazzmaster.deployment.Deployer;
import org.eiichiro.jazzmaster.deployment.Stage;
import org.eiichiro.jazzmaster.kernel.Kernel;
import org.eiichiro.jazzmaster.kernel.Property;
import org.eiichiro.jazzmaster.management.Configuration;
import org.eiichiro.jazzmaster.management.Manager;
import org.eiichiro.reverb.lang.ClassResolver;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Bootstrap} is a bootstrap process representation of Jazzmaster 
 * application. This class is launched by {@code Application}.
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Bootstrap {

	private Logger logger = LoggerFactory.getLogger(Bootstrap.class);
	
	private Stage stage;
	
	private Configuration configuration;
	
	private Kernel kernel;
	
	private Manager manager;
	
	private Deployer deployer;
	
	/**
	 * Constructs a new {@code Bootstrap} instance with the specified 
	 * deployment {@code Stage} and {@code Configuration}.
	 * 
	 * @param stage The deployment stage this Jazzmaster application runs.
	 * @param configuration The configuration for this Jazzmaster application.
	 */
	Bootstrap(Stage stage, Configuration configuration) {
		this.stage = stage;
		this.configuration = configuration;
	}
	
	/** Starts Jazzmaster application's bootstrap process. */
	public void bootstrap() {
		logger.info("Starting kernel");
		Kernel kernel = new Kernel();
		kernel.setGlobal(new Property<ClassResolver>(ClassResolver.class, new ClassResolver()));
		kernel.setGlobal(new Property<Stage>(Stage.class, stage));
		kernel.setGlobal(new Property<Configuration>(Configuration.class, configuration));
		kernel.setGlobal(new Property<Kernel>(Kernel.class, kernel));
		logger.trace("Starting manager");
		Manager manager = kernel.assemble(Manager.class).getInstance();
		kernel.setGlobal(new Property<Manager>(Manager.class, manager));
		logger.info("Manager started");
		logger.trace("Starting deployer");
		Deployer deployer = kernel.assemble(Deployer.class).getInstance();
		kernel.setGlobal(new Property<Deployer>(Deployer.class, deployer));
		logger.info("Deployer started");
		manager = kernel.assemble(manager).getInstance();
		manager.manage();
		deployer.deploy();
		logger.info("Kernel started");
		this.kernel = kernel;
		this.manager = manager;
		this.deployer = deployer;
	}

	public Kernel getKernel() {
		return kernel;
	}

	public Manager getManager() {
		return manager;
	}

	public Deployer getDeployer() {
		return deployer;
	}
	
}
