/*
 * Copyright (C) 2009-2012 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.bus;

import java.util.HashSet;
import java.util.Set;

import org.eiichiro.jazzmaster.kernel.Injected;
import org.eiichiro.jazzmaster.repository.Key;
import org.eiichiro.jazzmaster.repository.Repository;
import org.eiichiro.jazzmaster.service.Consumes;
import org.eiichiro.jazzmaster.service.Service;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Exposer} is holding service definition of the service that can be 
 * exposed to frontend and other services.
 * Detecting the declaration of service at frontend's field or service's field (
 * detecting {@code @Consumes} annotated field), Jazzmaster dependency injection 
 * mechanism would inject the service proxy to it. For service proxy 
 * construction, the bus requests the service definition to exposer with the 
 * {@code Key} of service definition that generated from frontend or service 
 * field's declaration. Exposer has the {@code Repository} reference in its 
 * field. When service definition is required form service proxy, exposer 
 * forwards the request to repository. If service definition cannot be found in 
 * repository or exposer isn't holding the definition that is the same of 
 * repository returns, exposer throws {@code NotExposedException}. Next, service 
 * proxy requests the service instance with the service definition. 
 * Exposer also forwards this request to repository and returns the instance 
 * that repository returns.
 *
 * @see Consumes
 * @see Key
 * @see Repository
 * @see NotExposedException
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Exposer {

	private Logger logger = LoggerFactory.getLogger(Exposer.class);
	
	private Set<Service> exposures = new HashSet<Service>();
	
	@Injected private Repository repository;
	
	/**
	 * Indicates whether the specified service should be expose or not.
	 * This exposer implementation assumes that the service can be expose, if 
	 * the service meets the following conditional: 
	 * <ul>
	 * <li>A service implementation class is declared as public.</li>
	 * <li>A service implementation class implements at least one interface.</li>
	 * </ul>
	 * 
	 * @param service The service definition to be exposed.
	 * @return The service definition has been exposed or not.
	 */
	public boolean expose(Service service) {
		Set<Class<?>> interfaces = service.getInterfaces();
		
		if (interfaces == null || interfaces.size() == 0) {
			logger.warn("Service [" + service.getName()
					+ "] is not exposed because of having no interface");
			return false;
		} else {
			exposures.add(service);
			logger.debug("Service [" + service.getName() + "] is exposed");
			return true;
		}
	}
	
	/**
	 * Returns the service definition related to the specified key.
	 * If related service is not found or not being held, this method throws an 
	 * {@code NotExposedException}.
	 * 
	 * @param key The key to get the service definition.
	 * @return The service definition related to the specified key.
	 * @see Repository#find(Key)
	 */
	public Service getService(Key<?> key) {
		Service service = repository.find(key);
		
		if (service == null) {
			throw new NotExposedException(
					"Service cannot be identified in repository; Key is ["
							+ key + "]");
		} else if (!exposures.contains(service)) {
			throw new NotExposedException("Service [" + service
					+ "] is not exposed");
		} else {
			return service;
		}
	}
	
	/**
	 * Returns the service instance related to the specified service definition.
	 * This method invocation is forwarded to repository's the same method.
	 * 
	 * @param service The service definition.
	 * @return The service instance.
	 * @see Repository#find(Service)
	 */
	public Object getService(Service service) {
		return repository.find(service);
	}

	public void setRepository(Repository repository) {
		this.repository = repository;
	}
	
}
