/*
 * Copyright (C) 2009-2012 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.bus;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * {@code Interceptor} is the annotation to declare service interceptor for 
 * Jazzmaster services invocation.
 * For interceptor declaration and interception of services, you need to use 
 * {@code Interception} annotation. And Jazzmaster interceptor supports only 
 * service <b>method</b> invocation and four types of advice ({@code Before}, 
 * {@code After}, {@code Around} and after {@code Throwing}) may be declared. 
 * For example: 
 * <pre>
 * package jazzmaster.hello;
 * 
 * import org.eiichiro.jazzmaster.bus.Around;
 * import org.eiichiro.jazzmaster.bus.Interceptor;
 * import org.eiichiro.jazzmaster.kernel.Injected;
 * import org.eiichiro.jazzmaster.service.Consumes;
 * import org.eiichiro.reverb.reflection.Invocation;
 * 
 * {@code @Logging} {@code @Interceptor}
 * public class LoggingInterceptor {
 * 
 *     {@code @Injected} {@code @Consumes} MyService myService;
 *     
 *     {@code @Around}
 *     public String trace(Invocation&lt;String&gt; invocation) throws Throwable {
 *         System.out.println();
 *         System.out.println("Hello#sayHello() start");
 *         System.out.println();
 *         String hello = invocation.proceed();
 *         System.out.println("Hello#sayHello() returns [" + hello + "]");
 *         System.out.println();
 *         System.out.println("Hello#sayHello() end");
 *         System.out.println();
 *         return hello;
 *     }
 * 
 *     public void setMyService(MyService myService) {
 *         this.myService = myService;
 *     }
 * 
 * }
 * </pre>
 * Service interface is: 
 * <pre>
 * package jazzmaster.hello;
 * 
 * public interface Hello {
 * 
 *     {@code @Logging} String sayHello();
 * 	
 * }
 * </pre>
 * Jazzmaster interceptor needs to {@code @Interceptor} anntation and advice 
 * method needs to one advice annotation. And interceptor is bound to intercepted 
 * service method by annotation. In this case, {@code @Logging} is used. The 
 * annotation for interception of service method needs to be annotated with 
 * {@code @Interception} and {@code RetentionPolicy#RUNTIME} retention.
 * {@code Interceptor} is managed as a singleton instance, and can be injected 
 * service references.
 * <pre>
 * package jazzmaster.hello;
 * 
 * import java.lang.annotation.ElementType;
 * import java.lang.annotation.Retention;
 * import java.lang.annotation.RetentionPolicy;
 * import java.lang.annotation.Target;
 * 
 * import org.eiichiro.jazzmaster.bus.Interception;
 * 
 * {@code @Retention(RetentionPolicy.RUNTIME)}
 * {@code @Target({ElementType.TYPE, ElementType.METHOD})}
 * {@code @Interception}
 * public {@code @interface} Logging {}
 * </pre>
 * 
 * @see Before
 * @see After
 * @see Around
 * @see Throwing
 * @see Bus
 * @see Interception
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@Documented
@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.TYPE)
public @interface Interceptor {}
