/*
 * Copyright (C) 2009-2012 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.frontend;

import java.lang.annotation.Annotation;
import java.util.HashSet;
import java.util.Set;

import static org.eiichiro.jazzmaster.Application.*;

import org.eiichiro.jazzmaster.bus.Bus;
import org.eiichiro.jazzmaster.deployment.Deployer;
import org.eiichiro.jazzmaster.deployment.Stage;
import org.eiichiro.jazzmaster.kernel.Binding;
import org.eiichiro.jazzmaster.kernel.Kernel;
import org.eiichiro.jazzmaster.kernel.Property;
import org.eiichiro.jazzmaster.management.Configuration;
import org.eiichiro.jazzmaster.management.Manager;
import org.eiichiro.jazzmaster.repository.Key;
import org.eiichiro.jazzmaster.repository.Repository;
import org.eiichiro.jazzmaster.service.Consumes;
import org.eiichiro.jazzmaster.service.Refers;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Injection} is the cross-cutting functional component that injects 
 * service dependencies into {@code Frontend} instance.
 * When a frontend instance is required by 
 * {@code Application#getFrontend(Class)} method or a frontend constructor is 
 * called, Jazzmaster instantiates this class and executes service dependencies 
 * injection. In injection process, this class disassembles all fields of 
 * frontend, specifies the service reference field ({@code @Consumes} annotated 
 * field) and gets binding annotation on the service reference field. Then, this 
 * class requires the service proxy from {@code Bus} with field declared type 
 * and binding annotations. If the service proxy is returned, this class injects 
 * it into the frontend instance field. For especially, {@code Kernel}, 
 * {@code Manager}, {@code Stage}, {@code Configuration}, {@code Bus}, 
 * {@code Repository} and {@code Deployer} may be injected into frontend 
 * instance directly by declaring the field that typed its class and named its 
 * class name (Of course, the first character must change to the small letter), 
 * no need to qualify the field with any binding annotations. By the way, 
 * Jazzmaster dependency injection does <b>NOT</b> support static field 
 * injection and preferentially supports "Setter Injection". If the target field 
 * does not have the setter method, the dependency injection attempts to set a 
 * referential component into the field directly.
 *
 * @see Consumes
 * @see Binding
 * @see Bus
 * @see Kernel
 * @see Manager
 * @see Stage
 * @see Configuration
 * @see Repository
 * @see Deployer
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@SuppressWarnings("deprecation")
public class Injection {

	/** Logger for this {@code Injection} instance hierarchy. */
	protected Logger logger = LoggerFactory.getLogger(getClass());
	
	public void inject(Object frontend) {
		Kernel kernel = getKernel();
		Manager manager = getManager();
		Bus bus = manager.getBus();
		bus.detect(frontend.getClass());
		Set<Property<?>> properties = new HashSet<Property<?>>();
		
		for (Property<?> property : kernel.disassemble(frontend).getProperties()) {
			Set<Annotation> annotations = property.getAnnotations();
			boolean consumes = false;
			
			for (Annotation annotation : annotations) {
				Class<? extends Annotation> type = annotation.annotationType();
				
				if (type.equals(Refers.class) || type.equals(Consumes.class)) {
					consumes = true;
					break;
				}
			}
			
			if (consumes) {
				Set<Annotation> bindings = getBindings(annotations);
				Key<?> key = newKey(property.getType(), bindings);
				logger.debug("Generated key [" + key + "] from property ["
						+ property + "]");
				Object service = bus.getServiceProxy(key);
				
				if (service == null) {
					logger.warn("Service proxy is [" + service
							+ "]; generated key for looking up is [" + key
							+ "]");
				} else {
					properties.add(newProperty(property.getName(), property
							.getType(), service));
				}
			}
		}
		
		kernel.assemble(frontend).with(properties).getInstance();
		logger.trace("Injected services into frontend [" + frontend.getClass()
				+ "]");
	}
	
	private Set<Annotation> getBindings(Set<Annotation> annotations) {
		Set<Annotation> bindings = new HashSet<Annotation>();
		
		for (Annotation annotation : annotations) {
			if (annotation.annotationType().getAnnotation(Binding.class) != null) {
				bindings.add(annotation);
			}
		}
		
		return bindings;
	}
	
	private <T> Key<T> newKey(Class<T> type, Set<Annotation> bindings) {
		return new Key<T>(type, bindings);
	}
	
	@SuppressWarnings("unchecked")
	private <T> Property<T> newProperty(String name, Class<T> type, Object value) {
		return new Property<T>(name, type, (T) value);
	}
	
}
