/*
 * Copyright (C) 2009-2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.kernel;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.ClassUtils;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Disassembler} disassembles the specified instance or class, and then 
 * abstracts its properties and annotations.
 * Using this class, you inspect what properties and annotations the specified 
 * component has. And reconstructing disassembled properties, you can reassemble 
 * a new component with them.
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Disassembler<T> {

	private Logger logger = LoggerFactory.getLogger(Disassembler.class);
	
	private Class<?> clazz;
	
	private T instance;
	
	private Set<Property<?>> properties = new HashSet<Property<?>>();
	
	private Set<Annotation> annotations = new HashSet<Annotation>();
	
	private boolean disassembled = false;
	
	/**
	 * Constructs a new {@code Disassembler} instance with the specified 
	 * {@code Class}.
	 * 
	 * @param clazz The class to be disassembled.
	 */
	public Disassembler(Class<T> clazz) {
		if (clazz == null) {
			throw new IllegalArgumentException("'clazz' must not be [" + clazz
					+ "]");
		}
		
		this.clazz = clazz;
	}
	
	/**
	 * Constructs a new {@code Disassembler} instance with the specified 
	 * instance.
	 * 
	 * @param instance The instance to be disassembled.
	 */
	@SuppressWarnings("unchecked")
	public Disassembler(T instance) {
		if (instance == null) {
			throw new IllegalArgumentException("'instance' must not be ["
					+ instance + "]");
		}
		
		clazz = (Class<T>) instance.getClass();
		this.instance = instance;
	}
	
	@SuppressWarnings("unchecked")
	public void disassemble() {
		Set<Property<?>> properties = new HashSet<Property<?>>();
		List<Field> fields = new ArrayList<Field>();
		fields.addAll(Arrays.asList(clazz.getDeclaredFields()));
		List<Class<?>> superclasses = ClassUtils.getAllSuperclasses(clazz);
		
		for (Class<?> superclass : superclasses) {
			fields.addAll(Arrays.asList(superclass.getDeclaredFields()));
		}
		
		for (Field field : fields) {
			Object value = null;
			
			try {
				if (field.getModifiers() == Modifier.STATIC) {
					value = PropertyUtils.getProperty(clazz, field.getName());
				} else if (instance != null) {
					value = PropertyUtils.getProperty(instance, field.getName());
				}
				
			} catch (IllegalAccessException e) {
				logger.warn(e.getMessage(), e);
			} catch (InvocationTargetException e) {
				logger.warn(e.getMessage(), e);
			} catch (NoSuchMethodException e) {
				logger.debug("Cannot find getter method for field [" + field
						+ "]: " + e.getMessage());
				field.setAccessible(true);
				
				try {
					if (field.getModifiers() == Modifier.STATIC) {
						value = field.get(clazz);
					} else if (instance != null) {
						value = field.get(instance);
					}
					
				} catch (IllegalArgumentException e1) {
					logger.warn(e.getMessage(), e);
				} catch (IllegalAccessException e1) {
					logger.warn(e.getMessage(), e);
				}
			}
			
			properties.add(newProperty(field.getName(), field.getType(), value, getAnnotations(field)));
		}
		
		Set<Annotation> annotations = new HashSet<Annotation>();
		
		for (Annotation annotation : clazz.getAnnotations()) {
			annotations.add(annotation);
		}
		
		this.properties = properties;
		this.annotations = annotations;
		disassembled = true;
	}
	
	/**
	 * Returns the abstracted properties.
	 * If {@code #disassemble()} has not been invoked yet, this method invokes 
	 * it and returns the abstracted properties.
	 * 
	 * @return The abstracted properties.
	 */
	public Set<Property<?>> getProperties() {
		if (!disassembled) {
			disassemble();
		}
		
		return properties;
	}
	
	/**
	 * Returns the abstracted class annotations.
	 * If {@code #disassemble()} has not been invoked yet, this method invokes 
	 * it and returns the abstracted annotations.
	 * 
	 * @return The abstracted annotations.
	 */
	public Set<Annotation> getAnnotations() {
		if (!disassembled) {
			disassemble();
		}
		
		return annotations;
	}
	
	@SuppressWarnings("unchecked")
	private <P> Property<P> newProperty(String name, Class<P> type, Object value, Set<Annotation> annotations) {
		return new Property<P>(name, type, (P) value, annotations);
	}
	
	private Set<Annotation> getAnnotations(Field field) {
		Set<Annotation> annotations = new HashSet<Annotation>();
		
		for (Annotation annotation : field.getAnnotations()) {
			annotations.add(annotation);
		}
		
		return annotations;
	}
	
}
