/*
 * Copyright (C) 2009-2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.management;

/**
 * {@code Configuration} holds and provides multiple configuration values 
 * to your application at runtime.
 * If you bootstrap your application with this class implementation instance, 
 * the instance may be populated into your custom components by Jazzmaster 
 * dependency injection. So that, you can configure them with this interface.
 * For example: 
 * <pre>
 * package jazzmaster.hello;
 * 
 * import org.eiichiro.jazzmaster.kernel.Injected;
 * import org.eiichiro.jazzmaster.kernel.Kernel;
 * import org.eiichiro.jazzmaster.kernel.Property;
 * import org.eiichiro.jazzmaster.management.Configuration;
 * import org.eiichiro.jazzmaster.service.Implementation;
 * 
 * {@code @Implementation}
 * public class HelloImpl implements Hello {
 * 
 *     {@code @Injected} private Kernel kernel;
 * 	
 *     {@code @Injected} private Configuration configuration;
 * 	
 *     public Message sayHello() {
 *         String message = String.class.cast(configuration.get("hello.message"));
 *         return kernel.assemble(Message.class).with(new Property&lt;String&gt;(message)).getInstance();
 *     }
 * 
 *     public void setKernel(Kernel kernel) {
 *         this.kernel = kernel;
 *     }
 * 
 *     public void setConfiguration(Configuration configuration) {
 *         this.configuration = configuration;
 *     }
 * 
 * }
 * </pre>
 * Application bootstrap is: 
 * <pre>
 * package jazzmaster.hello;
 * 
 * import org.eiichiro.jazzmaster.Application;
 * 
 * public class HelloMain {
 * 
 *     public static void main(String[] args) {
 *         Application.bootstrap(new HelloConfiguration());
 *         HelloFrontend helloFrontend = new HelloFrontend();
 *         HelloFrontend frontend = Application.getFrontend(HelloFrontend.class);
 *         helloFrontend.sayHello();
 *         frontend.sayHello();
 *         Application.shutdown();
 *     }
 * 
 * }
 * </pre>
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public interface Configuration {

	/**
	 * Returns the configured value corresponding the specified key.
	 * 
	 * @param key The key to value.
	 * @return The configured value corresponding the specified key.
	 */
	public Object get(Object key);
	
	/**
	 * Sets the specified key and value.
	 * 
	 * @param key The key to value.
	 * @param value The value corresponding to the key.
	 */
	public void set(Object key, Object value);
	
}
