/*
 * Copyright (C) 2009-2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.management;

import java.util.concurrent.ConcurrentHashMap;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.eiichiro.jazzmaster.Application;
import org.eiichiro.jazzmaster.WebFilter;

/**
 * {@code Scope} is the context representation of managed Jazzmaster component 
 * is activated.
 * Jazzmaster components can be managed by corresponding to the scope that 
 * component specifies by itself, and each component instance is stored in the 
 * {@code Context} that the scope provides.
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public enum Scope {

	/**
	 * The none pseudo-scope.
	 * This is the pseudo-scope for the component which doesn't have its scope. 
	 * The {@code Context} provided by this scope never holds any components. 
	 * Method {@code get(Object key)} returns always {@code null} and method 
	 * {@code keySet()} returns always empty {@code Set}.
	 */
	NONE(new Context() {

		public void clear() {}

		public Object get(Object key) {
			return null;
		}

		public Set<Object> keySet() {
			return new HashSet<Object>(0);
		}

		public void put(Object key, Object value) {}
		
	}), 
	
	/**
	 * The dependent pseudo-scope.
	 * This is the pseudo-scope for the component which doesn't have its scope 
	 * and have the same scope of the other one that depends on this scoped 
	 * component (When the other component that depends on this scoped component 
	 * is deleted, this scoped one is deleted).
	 * The {@code Context} provided by this scope never holds any components. 
	 * Method {@code get(Object key)} returns always {@code null} and method 
	 * {@code keySet()} returns always empty {@code Set}.
	 */
	DEPENDENT(new Context() {

		public void clear() {}

		public Object get(Object key) {
			return null;
		}

		public Set<Object> keySet() {
			return new HashSet<Object>(0);
		}

		public void put(Object key, Object value) {}
		
	}), 
	
	/**
	 * The thread scope.
	 * This is the scope for the component which is instantiated for each 
	 * {@code Thread}. The {@code Context} provided by this scope holds 
	 * components in current {@code ThreadLocal} object.
	 */
	THREAD(new ThreadContext()), 
	
	/**
	 * The request scope.
	 * This is the scope for the component which is available while 
	 * {@code HttpServletRequest} is processed. The {@code Context} provided by 
	 * this scope holds components in current {@code HttpServletRequest} 
	 * attribute.
	 */
	REQUEST(new RequestContext()), 
	
	/**
	 * The session scope.
	 * This is the scope for the component which is available while 
	 * {@code HttpSession} is activated. The {@code Context} provided by this 
	 * scope holds components in current {@code HttpSession} attribute.
	 */
	SESSION(new SessionContext()), 
	
	/**
	 * The application scope.
	 * This is the scope for the component which is instantiated as "Singleton", 
	 * which only one instance exists in whole application. The {@code Context} 
	 * provided by this scope hold components in static {@code Map} object.
	 */
	APPLICATION(new ApplicationContext());
	
	private final Context context;
	
	Scope(Context context) {
		this.context = context;
	}
	
	/**
	 * @return the context
	 */
	public Context getContext() {
		return context;
	}
	
	private static class ThreadContext implements Context {

		private static ThreadLocal<Map<Object, Object>> threadLocal 
				= new ThreadLocal<Map<Object, Object>>() {
			
			public Map<Object, Object> initialValue() {
				return new ConcurrentHashMap<Object, Object>();
			}
			
		};
		
		public Object get(Object key) {
			return threadLocal.get().get(key);
		}
		
		public void put(Object key, Object value) {
			threadLocal.get().put(key, value);
		}
		
		public Set<Object> keySet() {
			return threadLocal.get().keySet();
		}
		
		public void clear() {
			threadLocal.get().clear();
		}
		
	}
	
	/** The key prefix to set the component to request attribute. */
	public static final String REQUEST_ATTRIBUTE 
			= "org.eiichiro.jazzmaster.request.";
	
	private static class RequestContext implements Context {
		
		public Object get(Object key) {
			checkWeb();
			HttpServletRequest httpServletRequest = WebFilter
					.getHttpServletRequest();

			synchronized (httpServletRequest) {
				return httpServletRequest.getAttribute(REQUEST_ATTRIBUTE
						+ key.toString());
			}
		}

		public Set<Object> keySet() {
			checkWeb();
			HttpServletRequest httpServletRequest = WebFilter
					.getHttpServletRequest();
			
			if (httpServletRequest == null) {
				return new HashSet<Object>(0);
			}
			
			synchronized (httpServletRequest) {
				@SuppressWarnings("unchecked")
				Enumeration<String> attributeNames = httpServletRequest
						.getAttributeNames();
				Set<Object> keySet = new HashSet<Object>();
				
				while (attributeNames.hasMoreElements()) {
					String nextElement = attributeNames.nextElement();
					
					if (nextElement.startsWith(REQUEST_ATTRIBUTE)) {
						keySet.add(nextElement.substring(REQUEST_ATTRIBUTE
								.length()));
					}
				}
				
				return keySet;
			}
		}

		public void put(Object key, Object value) {
			checkWeb();
			HttpServletRequest httpServletRequest = WebFilter
					.getHttpServletRequest();
			
			synchronized (httpServletRequest) {
				httpServletRequest.setAttribute(REQUEST_ATTRIBUTE
						+ key.toString(), value);
			}
		}

		public void clear() {
			checkWeb();
		}
		
	}
	
	/** The key prefix to set the component to session attribute. */
	public static final String SESSION_ATTRIBUTE 
			= "org.eiichiro.jazzmaster.session.";
	
	private static class SessionContext implements Context {
		
		public Object get(Object key) {
			checkWeb();
			HttpSession httpSession = WebFilter.getHttpSession();
			
			synchronized (httpSession) {
				return httpSession.getAttribute(SESSION_ATTRIBUTE
						+ key.toString());
			}
		}
		
		public Set<Object> keySet() {
			checkWeb();
			HttpSession httpSession = WebFilter.getHttpSession();
			
			if (httpSession == null) {
				return new HashSet<Object>(0);
			}
			
			synchronized (httpSession) {
				@SuppressWarnings("unchecked")
				Enumeration<String> attributeNames = httpSession
						.getAttributeNames();
				Set<Object> keySet = new HashSet<Object>();
				
				while (attributeNames.hasMoreElements()) {
					String nextElement = attributeNames.nextElement();
					
					if (nextElement.startsWith(SESSION_ATTRIBUTE)) {
						keySet.add(nextElement.substring(SESSION_ATTRIBUTE
								.length()));
					}
				}
				
				return keySet;
			}
		}

		public void put(Object key, Object value) {
			checkWeb();
			HttpSession httpSession = WebFilter.getHttpSession();
			
			synchronized (httpSession) {
				httpSession.setAttribute(SESSION_ATTRIBUTE + key.toString(),
						value);
			}
		}
		
		public void clear() {
			checkWeb();
		}
		
	}
	
	private static class ApplicationContext implements Context {
		
		private static Map<Object, Object> application 
				= new ConcurrentHashMap<Object, Object>();
		
		public Object get(Object key) {
			return application.get(key);
		}
		
		public void put(Object key, Object value) {
			application.put(key, value);
		}
		
		public Set<Object> keySet() {
			return application.keySet();
		}
		
		public void clear() {
			synchronized (application) {
				application.clear();
			}
		}
		
	}
	
	private static void checkWeb() {
		if (!Application.isWeb()) {
			throw new UnsupportedOperationException(
					"Web scope is not activated: The current application "
							+ "doesn't run in web application environment");
		}
	}
	
}
