/*
 * Copyright (C) 2009-2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.repository;

import static org.eiichiro.jazzmaster.management.Lifecycle.CONFIGURED;
import static org.eiichiro.jazzmaster.management.Lifecycle.DESTROYED;
import static org.eiichiro.jazzmaster.management.Lifecycle.STARTED;
import static org.eiichiro.jazzmaster.management.Lifecycle.STOPPED;
import static org.eiichiro.jazzmaster.service.Service.named;
import static org.eiichiro.jazzmaster.service.Service.typed;

import java.lang.annotation.Annotation;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.eiichiro.jazzmaster.bus.Bus;
import org.eiichiro.jazzmaster.deployment.Deployer;
import org.eiichiro.jazzmaster.deployment.Staged;
import org.eiichiro.jazzmaster.kernel.Binding;
import org.eiichiro.jazzmaster.kernel.Injected;
import org.eiichiro.jazzmaster.kernel.Kernel;
import org.eiichiro.jazzmaster.kernel.Property;
import org.eiichiro.jazzmaster.management.Context;
import org.eiichiro.jazzmaster.management.Lifecycle;
import org.eiichiro.jazzmaster.management.Manager;
import org.eiichiro.jazzmaster.management.On;
import org.eiichiro.jazzmaster.service.Consumes;
import org.eiichiro.jazzmaster.service.Named;
import org.eiichiro.jazzmaster.service.Provider;
import org.eiichiro.jazzmaster.service.Reference;
import org.eiichiro.jazzmaster.service.Refers;
import org.eiichiro.jazzmaster.service.Service;
import org.eiichiro.jazzmaster.service.Service.UnknownImplementation;
import org.eiichiro.jazzmaster.service.Typed;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Repository} is a built-in Jazzmaster service repository implementation.
 * {@code Repository} has the service definitions corresponding to the storing 
 * &amp; finding {@code Key} and manages service instances by working the 
 * {@code Manager} together. For getting a service instance, first of all, you 
 * need to get the service definition by invoking {@code #find(Key)} with the 
 * generated {@code Key}. At the next, invoking {@code #find(Service)} with the 
 * above service definition, you can get the service instance. When the service 
 * instance (proxy) is required from application frontend or the other services, 
 * {@code Bus} gets the service instance with the above way and generates the 
 * proxy.
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@SuppressWarnings("deprecation")
public class Repository {

	private Logger logger = LoggerFactory.getLogger(Repository.class);
	
	@Injected private Kernel kernel;
	
	@Injected private Manager manager;
	
	@Injected private Bus bus;
	
	private Map<Key<?>, Service> services = new ConcurrentHashMap<Key<?>, Service>();
	
	/**
	 * Stores the specified service definition ({@code Service}).
	 * Generates service definition storing & finding {@code Key}s from the 
	 * specified service definition and holds the specified service definition 
	 * with the generated {@code Key}s. This method is invoked by 
	 * {@code Deployer} at deployment process.
	 * 
	 * @param service The service definition to be stored.
	 * @see Service
	 * @see Key
	 * @see Deployer#deploy()
	 */
	public void store(Service service) {
		for (Key<?> key : generateKeys(service)) {
			services.put(key, service);
		}
		
		logger.debug("Service [" + service.getName() + "] is stored");
	}
	
	/**
	 * Returns service definition corresponding to the specified {@code Key}.
	 * This method compares every {@code Key} generated from stored service 
	 * definitions to the specified key. If matched service definition is only 
	 * one, this method returns it, if is none, returns {@code null}. If 
	 * multiple service definitions are matched, this method compares priority 
	 * of them as the following rule and returns high-priority service 
	 * definition. (However, multiple service definitions remain after do it, 
	 * this method returns {@code null})
	 * <ol>
	 * <li>
	 * (The service definition that) explicitly-specified deployment stage is 
	 * the same as current application is running on, and every binding 
	 * annotation completely equals to the specified key's binding annotations.
	 * </li>
	 * <li>
	 * Explicitly-specified deployment stage is the same as current application 
	 * is running on, and the specified key's all binding annotations are 
	 * contained in service definition's or the specified key has no binding 
	 * annotation.
	 * </li>
	 * <li>
	 * Deployment stage is not specified (by default), and every binding 
	 * annotation completely equals to the specified key's binding annotations.
	 * </li>
	 * <li>
	 * Deployment stage is not specified (by default), and the specified key's 
	 * all binding annotations are contained in service definition's or the 
	 * specified key has no binding annotation.
	 * </li>
	 * </ol>
	 * 
	 * @param key The key to find the corresponding service definition.
	 * @return The service definition corresponding to the specified {@code Key}.
	 * @see Key
	 * @see Key#contains(Key)
	 */
	public Service find(Key<?> key) {
		Set<Service> candidate = new HashSet<Service>();
		Map<Service, Key<?>> keys = new HashMap<Service, Key<?>>();
		
		for (Key<?> k : services.keySet()) {
			if (k.contains(key)) {
				Service service = services.get(k);
				candidate.add(service);
				keys.put(service, k);
			}
		}
		
		if (candidate.isEmpty()) {
			logger.warn("Service associated [" + key + "] was not found");
			return null;
		} else if (candidate.size() == 1) {
			return candidate.iterator().next();
		}
		
		Set<Service> filtered = filterByStage(candidate);
		
		if (filtered.size() == 1) {
			return filtered.iterator().next();
		} else if (filtered.size() > 1) {
			Set<Service> set = filterByBinding(key, filtered, keys);
			
			if (set.size() == 1) {
				return set.iterator().next();
			} else {
				logger.warn("Service associated [" + key + "] is duplicated");
				return null;
			}
		}
		
		Set<Service> set = filterByBinding(key, candidate, keys);
		
		if (set.size() == 1) {
			return set.iterator().next();
		} else {
			logger.warn("Service associated [" + key + "] is duplicated");
			return null;
		}
	}
	
	private Set<Service> filterByBinding(Key<?> key, Set<Service> candidate, Map<Service, Key<?>> keys) {
		Set<Service> services = new HashSet<Service>();
		Set<Annotation> bindings = key.getBindings();
		
		for (Service service : candidate) {
			Set<Annotation> set = new HashSet<Annotation>();
			
			for (Annotation binding : keys.get(service).getBindings()) {
				if (!binding.annotationType().equals(Named.class)
						&& !binding.annotationType().equals(Typed.class)) {
					set.add(binding);
				}
			}
			
			if (bindings.size() == set.size()) {
				services.add(service);
			}
		}
		
		return services;
	}
	
	private Set<Service> filterByStage(Set<Service> candidate) {
		Set<Service> services = new HashSet<Service>();
		
		for (Service service : candidate) {
			if (service.getClass().getName().equals("org.eiichiro.jazzmaster.deployment.Deployer$ServiceImpl")) {
				if (service.getImplementation().getAnnotation(Staged.class) != null) {
					services.add(service);
				}
				
				continue;
			}
			
			if (service.getClass().getAnnotation(Staged.class) != null) {
				services.add(service);
				continue;				
			}
			
			try {
				service.getClass().getDeclaredMethod("getStages");
				services.add(service);
			} catch (NoSuchMethodException e) {}
		}
		
		return services;
	}
	
	/**
	 * Returns service instance corresponding to the specified {@code Service} 
	 * definition.
	 * This method searches the {@code Context} of the specified service's 
	 * {@code Scope} with the specified {@code Service}. If the service instance 
	 * still has been managed in the scope's context, this method returns it 
	 * immediately. If the service instance is not found, this method assembles 
	 * a new service instance, then invokes the {@code Lifecycle#CONFIGURED} and 
	 * {@code Lifecycle#STARTED} lifecycle callback method, at last, binds it 
	 * into the context and returns it.
	 * 
	 * @param service The service definition for finding the service instance.
	 * @return The service instance corresponded to the specified service 
	 * definition.
	 * @see Service#getScope()
	 * @see Context
	 * @see Scope
	 * @see Manager#getContext(Scope)
	 * @see Manager#callback(Object, Lifecycle)
	 */
	public Object find(Service service) {
		Context context = service.getScope().getContext();
		Object object = context.get(service);
		
		if (object == null) {
			Object instance = assemble(service);
			
			if (instance != null) {
				manager.setCallback(instance);
				
				for (Lifecycle lifecycle : Lifecycle.values()) {
					manager.setCallback(instance, lifecycle, service.getCallbacks(lifecycle));
				}
				
				manager.callback(instance, CONFIGURED);
				context.put(service, instance);
				manager.callback(instance, STARTED);
				return instance;
			}
		}
		
		return object;
	}
	
	private Object assemble(Service service) {
		Provider<?> provider = kernel.assemble(service.getProvider()).getInstance();
		Object instance = provider.get();
		Set<Property<?>> properties = new HashSet<Property<?>>();
		
		for (Property<?> property : kernel.disassemble(instance).getProperties()) {
			Set<Annotation> annotations = property.getAnnotations();
			boolean consumes = false;
			
			for (Annotation annotation : annotations) {
				Class<? extends Annotation> type = annotation.annotationType();
				
				if (type.equals(Refers.class) || type.equals(Consumes.class)) {
					consumes = true;
					break;
				}
			}
			
			if (consumes) {
				Set<Annotation> bindings = getBindings(annotations);
				Object proxy = bus.getServiceProxy(newKey(property.getType(),
						bindings));
				
				if (proxy != null) {
					properties.add(newProperty(property.getName(), property
							.getType(), proxy));
				}
			}
		}
		
		for (Property<?> property : service.getProperties()) {
			Object value = property.getValue();
			
			if (value instanceof Reference) {
				Reference reference = Reference.class.cast(value);
				Set<Annotation> bindings = new HashSet<Annotation>();
				Typed typed = reference.getTyped();
				Named named = reference.getNamed();
				Object proxy;
				
				if (typed != null) {
					bindings.add(typed);
					proxy = bus.getServiceProxy(newKey(property.getType(),
							bindings));
				} else if (named != null) {
					bindings.add(named);
					proxy = bus.getServiceProxy(newKey(property.getType(),
							bindings));
				} else {
					proxy = bus.getServiceProxy(newKey(reference.getType(),
							reference.getBindings()));
				}
				
				if (proxy != null) {
					properties.add(newProperty(property.getName(), property
							.getType(), proxy));
				}
				
			} else {
				properties.add(newProperty(property.getName(), property
						.getType(), value));
			}
		}
		
		return kernel.assemble(instance).with(properties).getInstance();
	}
	
	private Set<Annotation> getBindings(Set<Annotation> annotations) {
		Set<Annotation> bindings = new HashSet<Annotation>();
		
		for (Annotation annotation : annotations) {
			if (annotation.annotationType().getAnnotation(Binding.class) != null) {
				bindings.add(annotation);
			}
		}
		
		return bindings;
	}
	
	/** Stops all service definitions. */
	@On(STOPPED)
	public void stop() {
		callback(STOPPED);
	}
	
	/** Destroys all service definitions. */
	@On(DESTROYED)
	public void destroy() {
		callback(DESTROYED);
		services.clear();
	}
	
	private void callback(Lifecycle lifecycle) {
		for (Service service : services.values()) {
			manager.callback(service, lifecycle);
		}
	}
	
	private Set<Key<?>> generateKeys(Service service) {
		logger.debug("Generating keys of service [" + service.getName() + "]");
		Set<Annotation> bindings = getBindings(kernel.disassemble(
				service.getClass()).getAnnotations());
		Class<?> implementation = service.getImplementation();
		
		if (implementation != null && !implementation.equals(UnknownImplementation.class)) {
			bindings.addAll(getBindings(kernel.disassemble(implementation)
					.getAnnotations()));
		}
		
		bindings.addAll(service.getBindings());
		bindings.add(typed(service.getClass()));
		bindings.add(named(service.getName()));
		Set<Key<?>> keys = new HashSet<Key<?>>();
		
		for (Class<?> clazz : service.getInterfaces()) {
			keys.add(newKey(clazz, bindings));
		}
		
		return keys;
	}
	
	private <T> Key<T> newKey(Class<T> type, Set<Annotation> bindings) {
		return new Key<T>(type, bindings);
	}
	
	@SuppressWarnings("unchecked")
	private <T> Property<T> newProperty(String name, Class<T> type, Object value) {
		return new Property<T>(name, type, (T) value);
	}
	
	public void setKernel(Kernel kernel) {
		this.kernel = kernel;
	}

	public void setManager(Manager manager) {
		this.manager = manager;
	}

	public void setBus(Bus bus) {
		this.bus = bus;
	}
	
}
