/*
 * Copyright (C) 2009-2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.service;

import java.lang.annotation.Annotation;

/**
 * {@code NamedImpl} is a built-in implementation of {@code Named} binding 
 * annotation.
 * {@code Repository} stores a service definition with the {@code Named} binding 
 * annotation. If the service definition or service implementation class is not 
 * annotated with {@code Named} binding annotation, repository generates a 
 * {@code Named} binding annotation from this class automatically.
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@SuppressWarnings("all")
class NamedImpl implements Named {

	private final String value;
	
	/**
	 * Constructs a new {@code NamedImpl} instance with the specified service 
	 * name.
	 * 
	 * @param value The service name.
	 */
	public NamedImpl(String value) {
		this.value = value;
	}
	
	/** Returns the service name referred by {@code Named} binding annotation. */
	public String value() {
		return value;
	}
	
	/** Returns {@code Named.class} as the annotation type.  */
	public Class<? extends Annotation> annotationType() {
		return Named.class;
	}
	
	/**
	 * Returns {@code true} if the specified object implements {@code Named} 
	 * interface and the specified {@code Named#value()} is equivalent to this 
	 * {@code #value()}.
	 * 
	 * @param obj The target object to compare with.
	 * @return {@code true} if the specified object is equivalent to this one.
	 * @see Annotation#equals(Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Named) {
			Named named = Named.class.cast(obj);
			
			if (value.equals(named.value())) {
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Returns the hash code of this instance.
	 * 
	 * @see Annotation#hashCode()
	 * @return The hash code of this instance.
	 */
	@Override
	public int hashCode() {
		return (127 * "value".hashCode()) ^ value.hashCode();
	}
	
	/**
	 * Returns the {@code String} representation of this instance.
	 * 
	 * @see Annotation#toString()
	 * @return The {@code String} representation of this instance.
	 */
	@Override
	public String toString() {
		return "@" + Named.class.getName() + "(value=" + value + ")";
	}
	
}
