/*
 * Copyright (C) 2009-2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.service;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashSet;
import java.util.Set;

import org.eiichiro.jazzmaster.bus.Bus;
import org.eiichiro.jazzmaster.kernel.Property;
import org.eiichiro.jazzmaster.repository.Key;
import org.eiichiro.jazzmaster.repository.Repository;

/**
 * {@code Reference} is the service reference representation to assemble 
 * service instance field by programmable service configuration.
 * you can configure your service with the programmable way by declaring 
 * {@code Service} extension class. For configuring service instance properties, 
 * you should override {@code Service#getProperties()} method and return the 
 * {@code Set} view of {@code Property}. If you want to configure the other 
 * service reference as a service instance field, you can do it by constructing 
 * a new {@code Property} instance with a new {@code Reference} instance (as a 
 * property's value). For example: 
 * <pre></pre>
 * If the {@code Property} which {@code Service#getProperties()} returns is the 
 * other service {@code Reference}, {@code Repository} gets the other service 
 * instance (proxy) from {@code Bus} with the {@code Key} generating from 
 * {@code Reference} and assembles service instance.
 *
 * @see Bus
 * @see Property
 * @see Key
 * @see Repository
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Reference {

	private Typed typed;
	
	private Named named;
	
	private Class<?> type;
	
	private Set<Annotation> bindings = new HashSet<Annotation>(0);
	
	/**
	 * Constructs a new {@code Reference} instance with the specified 
	 * {@code Typed} binding annotation.
	 * {@code Typed} binding annotation can be generated by using 
	 * {@code Service#typed(Class)}.
	 * 
	 * @param typed The {@code Typed} binding annotation to refer the other 
	 * service.
	 * @see Service#typed(Class)
	 * @see Typed
	 */
	public Reference(Typed typed) {
		this.typed = typed;
	}
	
	/**
	 * Constructs a new {@code Reference} instance with the specified 
	 * {@code Named} binding annotation.
	 * {@code Named} binding annotation can be generated by using 
	 * {@code Service#named(String)}.
	 * 
	 * @param named The {@code Named} binding annotation to refer the other 
	 * service.
	 * @see Service#named(String)
	 * @see Named
	 */
	public Reference(Named named) {
		this.named = named;
	}
	
	/**
	 * Constructs a new {@code Reference} instance with the specified type and 
	 * binding annotation types.
	 * The specified type must be the one of the service implementation class's 
	 * declaration hierarchy or the service interface class's declaration 
	 * hierarchy. This method generates binding annotation instances from the 
	 * specified binding annotation types. The specified binding annotation 
	 * types <b>MUST NOT</b> have its own attribute. If you want to specify the 
	 * binding annotation that has its own attribute, you need to instantiate 
	 * binding annotation instance yourself and invoke another constructor (
	 * {@code #Reference(Class, Set, Class...)}).
	 * 
	 * @param type The {@code Class} the service represents.
	 * @param bindingTypes The array of binding annotation {@code Class}
	 * @throws UnsupportedOperationException If the specified binding annotation 
	 * type has its own attribute.
	 */
	public Reference(Class<?> type, Class<? extends Annotation>... bindingTypes) {
		this.type = type;
		Set<Annotation> bindings = new HashSet<Annotation>(bindingTypes.length);
		
		for (Class<? extends Annotation> bindingType : bindingTypes) {
			bindings.add(newBinding(bindingType));
		}
		
		this.bindings = bindings;
	}
	
	/**
	 * Constructs a new {@code Reference} instance with the specified type, 
	 * binding annotation instances and binding annotation types.
	 * The specified type must be the one of the service implementation class's 
	 * declaration hierarchy or the service interface class's declaration 
	 * hierarchy. This method generates binding annotation instances from the 
	 * specified binding annotation types. The binding annotation types 
	 * <b>MUST NOT</b> have its own attribute. If you want to specify the 
	 * binding annotation that has its own attribute, you need to instantiate 
	 * binding annotation instance yourself.
	 * 
	 * @param type The {@code Class} the service represents.
	 * @param bindings The {@code Set} view of binding annotation instance 
	 * that the service represents.
	 * @param bindingTypes The array of binding annotation {@code Class}
	 * @throws UnsupportedOperationException If the specified binding annotation 
	 * type has its own attribute.
	 */
	public Reference(Class<?> type, Set<Annotation> bindings,
			Class<? extends Annotation>... bindingTypes) {
		this.type = type;
		
		for (Class<? extends Annotation> bindingType : bindingTypes) {
			bindings.add(newBinding(bindingType));
		}
		
		this.bindings = bindings;
	}
	
	@SuppressWarnings("unchecked")
	private <T extends Annotation> T newBinding(final Class<T> bindingType) {
		InvocationHandler invocationHandler = new InvocationHandler() {

			public Object invoke(Object proxy, Method method, Object[] args)
					throws Throwable {
				String name = method.getName();
				
				if (name.equals("equals")) {
					if (args[0] instanceof Annotation) {
						Annotation annotation = Annotation.class.cast(args[0]);
						return annotation.annotationType().equals(bindingType);
					}
					
					return false;
				} else if (name.equals("hashCode")) {
					return 0;
				} else if (name.equals("toString")) {
					return "@" + bindingType.getName();
				} else if (name.equals("annotationType")) {
					return bindingType;
				} else {
					throw new UnsupportedOperationException("'bindingTypes' " 
							+ "passed to org.eiichiro.jazzmaster.service.Reference " 
							+ "must not have its own attribute");
				}
			}
			
		};
		Class<?>[] bindingTypes = {bindingType};
		return (T) Proxy.newProxyInstance(Thread.currentThread()
				.getContextClassLoader(), bindingTypes, invocationHandler);
	}

	public Typed getTyped() {
		return typed;
	}

	public Named getNamed() {
		return named;
	}

	public Class<?> getType() {
		return type;
	}

	public Set<Annotation> getBindings() {
		return bindings;
	}
	
}
